
/*

   (c) 2000 by A Vision of Paradise


   This file is part of the game 'Gamma Patrol'.

   'Gamma Patrol' is SHAREWARE, please look at the files 'README' and
   'LICENCE' for further details.

   The source code is NOT OpenSource, GPL'd or what ever!
   You can use it for educational purposes, but you are not allowed
   to distribute modified or partial version of 'Gamma Patrol'.

*/

/* $Id: level.h,v 1.1 2000/10/11 20:04:05 pink Exp pink $ */

#ifndef _LEVEL_H_
#define _LEVEL_H_

#include "main.h"
#include "object.h"
#include "svector.h"
#include "sarray.h"

enum {
  LDL_NODE_INT = 0,
  LDL_NODE_STRING,
  LDL_NODE_OP,
  LDL_NODE_OBJECT,
  LDL_NODE_BONUS,
  LDL_NODE_MODIFIER,
  LDL_NODE_TRIGGER,
  LDL_NODE_WHILE,
  LDL_NODE_REPEAT,
  LDL_NODE_IF,
  LDL_NODE_DEFINE,
  LDL_NODE_SUB,
  LDL_NODE_CALL,
  LDL_NODE_ID
};

class ldl_node {
 private:
  Uint16 _type;
  ldl_node *_child;
  ldl_node *_next;

 public:
  ldl_node(Uint16 t) {
    _type = t;
    _next = NULL;
    _child = NULL;
  }

  ldl_node(const ldl_node &n) {
    _type = n._type;
    _next = n._next ? n._next->clone() : NULL;
    _child = n._child ? n._child->clone() : NULL;
  }

  virtual ~ldl_node() {
    delete _next;
    delete _child;
  }

  Uint16 type() const { return(_type); }
  ldl_node *next() const { return(_next); }
  void next(ldl_node *n) { _next = n; }
  ldl_node *child() const { return(_child); }
  void child(ldl_node *c) { _child = c; }

  virtual ldl_node *clone() const = 0;
  virtual void class_scan(class level *);
  virtual Sint32 eval(class level *) const;
  virtual void execute(class level *) const;
  virtual void dump() const;
};

class ldl_node_string : public ldl_node {
 private:
  char *_str;

 public:
  ldl_node_string(const char *s) : ldl_node(LDL_NODE_STRING) {
    _str = new char[strlen(s) + 1];
    strcpy(_str, s);
  }

  ldl_node_string(const ldl_node_string &n) : ldl_node(n) {
    _str = new char[strlen(n._str) + 1];
    strcpy(_str, n._str);
  }

  ~ldl_node_string() {
    delete _str;
  }

  char *str() const { return(_str); }

  ldl_node *clone() const {
    return( new ldl_node_string(*this) );
  }

  void dump() const;
};

class ldl_node_int : public ldl_node {
 private:
  Sint32 _val;

 public:
  ldl_node_int(Sint32 v) : ldl_node(LDL_NODE_INT) {
    _val = v;
  }

  ldl_node_int(const ldl_node_int &n) : ldl_node(n) {
    _val = n._val;
  }

  Sint32 val() const { return(_val); }

  ldl_node *clone() const {
    return( new ldl_node_int(*this) );
  }

  Sint32 eval(class level *) const { return(_val); }
  void dump() const;
};

class ldl_node_id : public ldl_node {
 public:
  ldl_node_id() : ldl_node(LDL_NODE_ID) {
  }

  ldl_node_id(const ldl_node_id &n) : ldl_node(n) {
  }

  ldl_node *clone() const {
    return( new ldl_node_id(*this) );
  }

  Sint32 eval(class level *) const;
  void dump() const;
};

enum {
  LDL_OP_ADD = 0,
  LDL_OP_SUB,
  LDL_OP_MUL,
  LDL_OP_DIV,
  LDL_OP_MOD,
  LDL_OP_NEG,
  LDL_OP_GT,
  LDL_OP_GE,
  LDL_OP_LT,
  LDL_OP_LE,
  LDL_OP_EQ,
  LDL_OP_NE
};

class ldl_node_op : public ldl_node {
 private:
  Uint16 _op;

 public:
  ldl_node_op(Uint16 o) : ldl_node(LDL_NODE_OP) {
    _op = o;
  }

  ldl_node_op(const ldl_node_op &n) : ldl_node(n) {
    _op = n._op;
  }

  Uint16 op() const { return(_op); }

  ldl_node *clone() const {
    return( new ldl_node_op(*this) );
  }

  Sint32 eval(class level *) const;
  void dump() const;
};

class ldl_node_object : public ldl_node {
 private:
  Uint16 _id;

 public:
  ldl_node_object() : ldl_node(LDL_NODE_OBJECT) {
    _id = 0;
  }

  ldl_node_object(const ldl_node_object &n) : ldl_node(n) {
    _id = n._id;
  }

  ldl_node *clone() const {
    return( new ldl_node_object(*this) );
  }

  void class_scan(class level *);
  void execute(class level *) const;
  void dump() const;
};

class ldl_node_modifier : public ldl_node {
 public:
  ldl_node_modifier() : ldl_node(LDL_NODE_MODIFIER) {
  }

  ldl_node_modifier(const ldl_node_modifier &n) : ldl_node(n) {
  }

  ldl_node *clone() const {
    return( new ldl_node_modifier(*this) );
  }

  void build_modifier_table(class level *, svector<object_modifier> &) const;
  void dump() const;
};

class ldl_node_bonus : public ldl_node {
 public:
  ldl_node_bonus() : ldl_node(LDL_NODE_BONUS) {
  }

  ldl_node_bonus(const ldl_node_bonus &n) : ldl_node(n) {
  }

  ldl_node *clone() const {
    return( new ldl_node_bonus(*this) );
  }

  void execute(class level *) const;
  void dump() const;
};

class ldl_node_trigger : public ldl_node {
 private:

 public:
  ldl_node_trigger() : ldl_node(LDL_NODE_TRIGGER) {
  }

  ldl_node_trigger(const ldl_node_trigger &n) : ldl_node(n) {
  }

  ldl_node *clone() const {
    return( new ldl_node_trigger(*this) );
  }

  void execute(class level *) const;
  void dump() const;
};

class ldl_node_trigger_type : public ldl_node {
 private:
  Uint16 _trig;

 public:
  ldl_node_trigger_type(Uint16 t) : ldl_node(LDL_NODE_TRIGGER) {
    _trig = t;
  }

  ldl_node_trigger_type(const ldl_node_trigger_type &n) : ldl_node(n) {
    _trig = n._trig;
  }

  Uint16 trig() const { return( _trig ); }

  ldl_node *clone() const {
    return( new ldl_node_trigger_type(*this) );
  }

  void dump() const;
};

class ldl_node_sub : public ldl_node {
 private:

 public:
  ldl_node_sub() : ldl_node(LDL_NODE_SUB) {
  }

  ldl_node_sub(const ldl_node_sub &n) : ldl_node(n) {
  }

  ldl_node *clone() const {
    return( new ldl_node_sub(*this) );
  }

  void execute(class level *) const;
  void dump() const;
};

class ldl_node_call : public ldl_node {
 private:

 public:
  ldl_node_call() : ldl_node(LDL_NODE_CALL) {
  }

  ldl_node_call(const ldl_node_call &n) : ldl_node(n) {
  }

  ldl_node *clone() const {
    return( new ldl_node_call(*this) );
  }

  void execute(class level *) const;
  void dump() const;
};

class ldl_node_while : public ldl_node {
 private:

 public:
  ldl_node_while() : ldl_node(LDL_NODE_WHILE) {
  }

  ldl_node_while(const ldl_node_while &n) : ldl_node(n) {
  }

  ldl_node *clone() const {
    return( new ldl_node_while(*this) );
  }

  void execute(class level *) const;
  void dump() const;
};

class ldl_node_repeat : public ldl_node {
 private:

 public:
  ldl_node_repeat() : ldl_node(LDL_NODE_REPEAT) {
  }

  ldl_node_repeat(const ldl_node_repeat &n) : ldl_node(n) {
  }

  ldl_node *clone() const {
    return( new ldl_node_repeat(*this) );
  }

  void execute(class level *) const;
  void dump() const;
};

class ldl_node_if : public ldl_node {
 private:

 public:
  ldl_node_if() : ldl_node(LDL_NODE_IF) {
  }

  ldl_node_if(const ldl_node_if &n) : ldl_node(n) {
  }

  ldl_node *clone() const {
    return( new ldl_node_if(*this) );
  }

  void execute(class level *) const;
  void dump() const;
};

class ldl_node_define : public ldl_node {
 private:

 public:
  ldl_node_define() : ldl_node(LDL_NODE_DEFINE) {
  }

  ldl_node_define(const ldl_node_define &n) : ldl_node(n) {
  }

  ldl_node *clone() const {
    return( new ldl_node_define(*this) );
  }

  void execute(class level *) const;
  void dump() const;
};

enum {
  LDL_TRIGGER_POSITION,
  LDL_TRIGGER_RELATIVE,
  LDL_TRIGGER_SIGNAL,
  LDL_TRIGGER_TIME
};

class ldl_trigger {
 private:
  Uint16 _type;
  ldl_node *_code;
  Sint32 _xpos, _ypos;
  Sint32 _dist;
  Uint32 _signal;
  Uint32 _time;

 public:
  ldl_trigger(Uint16 t, ldl_node *c) {
    _type = t;
    _code = c;
    _xpos = 0;
    _ypos = 0;
    _dist = 0;
    _signal = 0;
    _time = 0;
  }

  ~ldl_trigger() {
    delete _code;
  }

  Uint16 type() const { return( _type ); }
  ldl_node *code() const { return( _code ); }
  Sint32 xpos() const { return( _xpos ); }
  void xpos(Sint32 v) { _xpos = v; }
  Sint32 ypos() const { return( _ypos ); }
  void ypos(Sint32 v) { _ypos = v; }
  Sint32 dist() const { return( _dist ); }
  void dist(Sint32 v) { _dist = v; }
  Uint32 signal() const { return( _signal ); }
  void signal(Uint32 v) { _signal = v; }
  Uint32 time() const { return( _time ); }
  void time(Uint32 v) { _time = v; }
};

class level {
 protected:
  Uint16 _song;

  svector<Uint16> _classes;

  Sint32 _width;
  Sint32 _height;
  Sint32 _yvel;

  sarray<ldl_trigger> _triggers;
  sarray<ldl_trigger> _added_triggers;

  ldl_node *_root;
  bool load(char *);

 public:
  level();
  virtual ~level();

  virtual bool setup();
  virtual void cleanup();

  Uint16 song() const { return( _song ); }

 private:
  struct var { char *name; Sint32 val; };
  sarray<var> _vars;
  void cleanup_variables();

 public:
  void set_variable(const char *, Sint32);
  Sint32 get_variable(const char *);

 private:
  struct sub { char *name; ldl_node *root; };
  sarray<sub> _subs;
  void cleanup_subs();

 public:
  void set_sub(const char *, ldl_node *);
  ldl_node *get_sub(const char *);

  Uint16 register_class(char *);
  void add_trigger(ldl_trigger *t) {
    _added_triggers += t;
  }

  Sint32 width() const { return( _width ); }
  Sint32 height() const { return( _height ); }
  Sint32 yvel() const { return( _yvel ); }

  virtual void render_background() = 0;
  void check_triggers();
};

#endif

