
/*

   (c) 2000 by A Vision of Paradise


   This file is part of the game 'Gamma Patrol'.

   'Gamma Patrol' is SHAREWARE, please look at the files 'README' and
   'LICENCE' for further details.

   The source code is NOT OpenSource, GPL'd or what ever!
   You can use it for educational purposes, but you are not allowed
   to distribute modified or partial version of 'Gamma Patrol'.

*/

/* $Id: surface.cc,v 1.1 2000/10/11 20:04:06 pink Exp pink $ */

#include <stdlib.h>
#include <memory.h>
#include <ctype.h>
#include <zlib.h>
#include <SDL/SDL_image.h>

#include "surface.h"
#include "font.h"

surface::surface() {
  _width = 0;
  _height = 0;
  _frames = 1;

  _surface = NULL;
  _bound = NULL;

  _last_frame = 0;
  _current_frame = 0;
}

surface::surface(SDL_Surface *s) {
  _width = s->w;
  _height = s->h;
  _frames = 1;

  _surface = new SDL_Surface *[1];
  _surface[0] = s;
  _bound = new rectangle[1];
  _bound[0] = rectangle(0, 0, _width, _height);

  _last_frame = 0;

  _clip_box.x(0);
  _clip_box.y(0);
  _clip_box.w(_width);
  _clip_box.h(_height);

  _current_frame = 0;
}

surface::surface(Uint16 w, Uint16 h, Uint16 f) {
  _width = w;
  _height = h;
  _frames = f;

  _bound = new rectangle[_frames];
  _surface = new SDL_Surface *[_frames];
  memset(_surface, 0, _frames * sizeof(SDL_Surface *));

  _last_frame = 0;

  _clip_box.x(0);
  _clip_box.y(0);
  _clip_box.w(_width);
  _clip_box.h(_height);

  _current_frame = 0;
}

surface::~surface() {
  for ( Uint16 i = 0 ; i < _frames ; ++i )
    if ( _surface[i] )
      SDL_FreeSurface(_surface[i]);

  delete _surface;
  delete _bound;
}

void surface::add_sdl_surface(const rectangle &r, SDL_Surface *s) {
  if ( _last_frame != _frames ) {
    _bound[_last_frame] = r;
    _surface[_last_frame] = s;
    _last_frame++;
  }
}

Uint32 surface::memory() const {
  Uint32 m = 0;

  for ( Uint16 i = 0 ; i < _frames ; ++i ) {
    if ( _surface[i] )
      m += _surface[i]->pitch * _surface[i]->h;
  }

  return( m );
}

surface *surface::load_image(const char *path) {
  surface *result = NULL;

  gzFile fp = gzopen(path, "rb");
  if ( fp ) {
    SDL_Surface *surf = NULL;
    
    char id_str[8];

    if ( gzread(fp, id_str, 8) != 8 ) {
      printf("Unable to read magic number from '%s'.\n", path);
      goto done;
    }

    if ( strncmp(id_str, "AVOP/IMG", 8) == 0 ) {
      long width, height, frames, type;
      if ( (gzread(fp, &width, sizeof(long)) != sizeof(long)) ||
	   (gzread(fp, &height, sizeof(long)) != sizeof(long)) ||
	   (gzread(fp, &frames, sizeof(long)) != sizeof(long)) ||
	   (gzread(fp, &type, sizeof(long)) != sizeof(long)) ) {
	printf("I/O error while reading '%s'.\n", path);
	goto done;
      }

      result = new surface(width, height, frames);

      for ( int i = 0 ; i < frames ; i++ ) {
	long x, y, w, h;
	if ( (gzread(fp, &x, sizeof(long)) != sizeof(long)) ||
	     (gzread(fp, &y, sizeof(long)) != sizeof(long)) ||
	     (gzread(fp, &w, sizeof(long)) != sizeof(long)) ||
	     (gzread(fp, &h, sizeof(long)) != sizeof(long)) ) {
	  printf("I/O error while reading '%s'.\n", path);
	  goto done;
	}
	
	SDL_Surface *surf = NULL;

	switch ( type ) {
	case 3 :
	  surf = SDL_AllocSurface(SDL_SWSURFACE | SDL_SRCALPHA, w, h, 32,
				  0xFF000000, 0x00FF0000, 0x0000FF00, 0x000000FF);
	  if ( surf ) {
	    if ( !SDL_MUSTLOCK(surf) || (SDL_LockSurface(surf) >= 0) ) {
	      Uint8 *buf = new Uint8[4 * w];
	    
	      for ( int y = 0 ; y < h ; ++y ) {
		Uint32 *dest = (Uint32 *)((Uint8 *)surf->pixels + y*surf->pitch);
		Uint8 *src = buf;

		if ( gzread(fp, buf, 4 * w) != 4 * w ) {
		  printf("I/O error while reading '%s'.\n", path);
		  goto done;
		}

		for ( int x = 0 ; x < w ; ++x ) {
#if ( SDL_VERSIONNUM(SDL_MAJOR_VERSION, SDL_MINOR_VERSION, SDL_PATCHLEVEL) < 1105 )
		  *dest++ = ((Uint32)src[0] << 24) | ((Uint32)src[1] <<  16) | ((Uint32)src[2] << 8) | (Uint32)src[3];
#else
		  /* alpha value flipped with SDL 1.1.5 */
		  *dest++ = ((Uint32)src[0] << 24) | ((Uint32)src[1] <<  16) | ((Uint32)src[2] << 8) | (Uint32)(255-src[3]);
#endif
		  src += 4;
		}
	      }
	
	      delete buf;

	      if ( SDL_MUSTLOCK(surf) )
		SDL_UnlockSurface(surf);

	      SDL_SetAlpha(surf, SDL_RLEACCEL, 0);
#if ( SDL_VERSIONNUM(SDL_MAJOR_VERSION, SDL_MINOR_VERSION, SDL_PATCHLEVEL) >= 1105 )
	      SDL_Surface *tmp = SDL_DisplayFormatAlpha(surf);
	      if ( tmp ) {
		SDL_FreeSurface(surf);
		surf = tmp;
	      }
#endif
	    }
	  }
	  else
	    printf("Unable to allocate %ldx%ld surface for '%s'.\n", width, height, path);
	  break;
      
	case 2 :
	  surf = SDL_AllocSurface(SDL_SWSURFACE, w, h, 24,
				  0x00FF0000, 0x0000FF00, 0x000000FF, 0x00000000);
	  if ( surf ) {
	    if ( !SDL_MUSTLOCK(surf) || (SDL_LockSurface(surf) >= 0) ) {
	      Uint8 *buf = new Uint8[3 * w];
	    
	      for ( int y = 0 ; y < h ; ++y ) {
		Uint8 *dest = (Uint8 *)surf->pixels + y*surf->pitch;
		Uint8 *src = buf;

		if ( gzread(fp, buf, 3 * w) != 3 * w ) {
		  printf("I/O error while reading '%s'.\n", path);
		  goto done;
		}

		for ( int x = 0 ; x < w ; ++x ) {
		  *(dest + surf->format->Rshift/8) = src[0]; 
		  *(dest + surf->format->Gshift/8) = src[1];
		  *(dest + surf->format->Bshift/8) = src[2];
		  dest += 3;
		  src += 3;
		}
	      }

	      delete buf;

	      if ( SDL_MUSTLOCK(surf) )
		SDL_UnlockSurface(surf);
	    }
	  }
	  else
	    printf("Unable to allocate %ldx%ld surface for '%s'.\n", width, height, path);
	  break;
      
	default:
	  printf("Unknown image format in '%s'.\n", path);
	  break;
	}

	if ( surf )
	  result->add_sdl_surface(rectangle(x, y, w, h), surf);
	else
	  goto done;
      }
    }
    else if ( strncmp(id_str, "AVOP/IM2", 8) == 0 ) {
      long width, height, frames, type;
      if ( (gzread(fp, &width, sizeof(long)) != sizeof(long)) ||
	   (gzread(fp, &height, sizeof(long)) != sizeof(long)) ||
	   (gzread(fp, &frames, sizeof(long)) != sizeof(long)) ||
	   (gzread(fp, &type, sizeof(long)) != sizeof(long)) ) {
	printf("I/O error while reading '%s'.\n", path);
	goto done;
      }

      result = new surface(width, height, frames);

      for ( int i = 0 ; i < frames ; i++ ) {
	long x, y, w, h;
	short t;
	if ( (gzread(fp, &x, sizeof(long)) != sizeof(long)) ||
	     (gzread(fp, &y, sizeof(long)) != sizeof(long)) ||
	     (gzread(fp, &w, sizeof(long)) != sizeof(long)) ||
	     (gzread(fp, &h, sizeof(long)) != sizeof(long)) ||
	     (gzread(fp, &t, sizeof(short)) != sizeof(short)) ) {
	  printf("I/O error while reading '%s'.\n", path);
	  goto done;
	}

	SDL_Surface *surf = NULL;

	switch ( t ) {
	case 0 :
	  switch ( type ) {
	  case 3 :
	    surf = SDL_AllocSurface(SDL_SWSURFACE | SDL_SRCALPHA, w, h, 32,
				    0xFF000000, 0x00FF0000, 0x0000FF00, 0x000000FF);
	    if ( surf ) {
	      if ( !SDL_MUSTLOCK(surf) || (SDL_LockSurface(surf) >= 0) ) {
		Uint8 *buf = new Uint8[4 * w];
	    
		for ( int y = 0 ; y < h ; ++y ) {
		  Uint32 *dest = (Uint32 *)((Uint8 *)surf->pixels + y*surf->pitch);
		  Uint8 *src = buf;

		  if ( gzread(fp, buf, 4 * w) != 4 * w ) {
		    printf("I/O error while reading '%s'.\n", path);
		    goto done;
		  }

		  for ( int x = 0 ; x < w ; ++x ) {
#if ( SDL_VERSIONNUM(SDL_MAJOR_VERSION, SDL_MINOR_VERSION, SDL_PATCHLEVEL) < 1105 )
		    *dest++ = ((Uint32)src[0] << 24) | ((Uint32)src[1] <<  16) | ((Uint32)src[2] << 8) | (Uint32)src[3];
#else
		    /* alpha value flipped with SDL 1.1.5 */
		    *dest++ = ((Uint32)src[0] << 24) | ((Uint32)src[1] <<  16) | ((Uint32)src[2] << 8) | (Uint32)(255-src[3]);
#endif
		    src += 4;
		  }
		}
	
		delete buf;

		if ( SDL_MUSTLOCK(surf) )
		  SDL_UnlockSurface(surf);

		SDL_SetAlpha(surf, SDL_RLEACCEL, 0);
#if ( SDL_VERSIONNUM(SDL_MAJOR_VERSION, SDL_MINOR_VERSION, SDL_PATCHLEVEL) >= 1105 )
		SDL_Surface *tmp = SDL_DisplayFormatAlpha(surf);
		if ( tmp ) {
		  SDL_FreeSurface(surf);
		  surf = tmp;
		}
#endif
	      }
	    }
	    else
	      printf("Unable to allocate %ldx%ld surface for '%s', frame %d.\n", w, h, path, i);
	    break;
      
	  case 2 :
	    surf = SDL_AllocSurface(SDL_SWSURFACE, w, h, 24,
				    0x00FF0000, 0x0000FF00, 0x000000FF, 0x00000000);
	    if ( surf ) {
	      if ( !SDL_MUSTLOCK(surf) || (SDL_LockSurface(surf) >= 0) ) {
		Uint8 *buf = new Uint8[3 * w];
	    
		for ( int y = 0 ; y < h ; ++y ) {
		  Uint8 *dest = (Uint8 *)surf->pixels + y*surf->pitch;
		  Uint8 *src = buf;

		  if ( gzread(fp, buf, 3 * w) != 3 * w ) {
		    printf("I/O error while reading '%s'.\n", path);
		    goto done;
		  }

		  for ( int x = 0 ; x < w ; ++x ) {
		    *(dest + surf->format->Rshift/8) = src[0]; 
		    *(dest + surf->format->Gshift/8) = src[1];
		    *(dest + surf->format->Bshift/8) = src[2];
		    dest += 3;
		    src += 3;
		  }
		}

		delete buf;

		if ( SDL_MUSTLOCK(surf) )
		  SDL_UnlockSurface(surf);
	      }
	    }
	    else
	      printf("Unable to allocate %ldx%ld surface for '%s', frame %d.\n", w, h, path, i);
	    break;
      
	  default:
	    printf("Unknown image format in '%s', frame %d.\n", path, i);
	    break;
	  }
	  break;

	case 1 :
	  switch ( type ) {
	  case 2 :
	  case 3 :
	    surf = SDL_AllocSurface(((type == 3) ? SDL_SRCCOLORKEY : SDL_SWSURFACE), w, h, 8,
				    0x00000000, 0x00000000, 0x00000000, 0x00000000);

	    if ( surf ) {
	      extern bool enable_debug;
	      if ( enable_debug && (surf->flags & SDL_HWSURFACE) ) {
		printf("'%s', frame %d is hw-surface.\n", path, i);
	      }

	      Uint16 cols;

	      if ( gzread(fp, &cols, sizeof(Uint16)) != sizeof(Uint16) ) {
		printf("I/O error while reading '%s'.\n", path);
		goto done;
	      }

	      Uint8 key = 0;

	      for ( Uint16 c = 0 ; c < cols ; ++c ) {
		Uint8 rgb[3];

		if ( gzread(fp, rgb, sizeof(rgb)) != sizeof(rgb) ) {
		  printf("I/O error while reading '%s'.\n", path);
		  goto done;
		}

		if ( (rgb[0] == 0) && (rgb[1] == 0) && (rgb[2] == 0) )
		  key = c;

		SDL_Color col;
		col.r = rgb[0];
		col.g = rgb[1];
		col.b = rgb[2];
		SDL_SetColors(surf, &col, c, 1);
	      }

	      if ( !SDL_MUSTLOCK(surf) || (SDL_LockSurface(surf) >= 0) ) {
		for ( int y = 0 ; y < h ; ++y ) {
		  Uint8 *dest = (Uint8 *)surf->pixels + y*surf->pitch;

		  if ( gzread(fp, dest, w) != w ) {
		    printf("I/O error while reading '%s'.\n", path);
		    goto done;
		  }
		}

		if ( SDL_MUSTLOCK(surf) )
		  SDL_UnlockSurface(surf);

		if ( type == 3 ) {
		  SDL_SetColorKey(surf, SDL_RLEACCEL | SDL_SRCCOLORKEY, key);
		  SDL_Surface *tmp = SDL_DisplayFormat(surf);
		  if ( tmp ) {
		    SDL_FreeSurface(surf);
		    surf = tmp;
		  }
		}
	      }
	    }
	    else
	      printf("Unable to allocate %ldx%ld surface for '%s', frame %d.\n", w, h, path, i);
	    break;
      
	  default:
	    printf("Unknown image format in '%s', frame %d.\n", path, i);
	    break; 
	  }
	  break;
	}

	if ( surf )
	  result->add_sdl_surface(rectangle(x, y, w, h), surf);
	else
	  goto done;
      }
    }
  
  done:
    gzclose(fp);
  }

  if ( result == NULL ) {
    SDL_Surface *surf = IMG_Load(path);
    if ( surf )
      result = new surface(surf);
    else
      printf("Unable to load image '%s':  %s\n", path, SDL_GetError());
  }

  return( result );
}

surface *surface::load_image(const char **path_list, const char **suffix_list, const char *name) {
  surface *result = NULL;

  for ( const char **path = path_list ; !result && *path ; ++path ) {
    for ( const char **suffix = suffix_list ; !result && *suffix ; ++suffix ) {
      char buf[256];
      sprintf(buf, "%s%s%s", *path, name, *suffix);
      result = load_image(buf);
    }
  }

  return( result );
}

void surface::display_format() {
  for ( Uint16 i = 0 ; i < _frames ; ++i ) {
    if ( _surface[i] ) {
      SDL_Surface *tmp = SDL_DisplayFormat(_surface[i]);
      if ( tmp ) {
	SDL_FreeSurface(_surface[i]);
	_surface[i] = tmp;
      }
    }
  }
}

rectangle surface::get_bound(const rectangle &r, Uint16 fr) {
  SDL_Surface *_s = _surface[fr];
  if ( !_s )
    return(rectangle(0, 0, 0, 0));

  Sint16 x_min = r.x() + r.w() - 1;
  Sint16 x_max = r.x();
  Sint16 y_min = r.y() + r.h() - 1;
  Sint16 y_max = r.y();

  switch ( _s->format->BytesPerPixel ) {
  case 2 :
    if ( lock(fr) ) {
      for ( int y = r.y() ; y < r.y() + r.h() ; ++y ) {
	Uint16 *row = (Uint16 *)((Uint8 *)_s->pixels + y * _s->pitch + r.x() * 2);

	for ( int x = r.x() ; x < r.x() + r.w() ; ++x ) {
	  bool hit = false;

	  if ( (*row++) & _s->format->Amask != 0 )
	    hit = true;

	  if ( hit ) {
	    if ( x < x_min )
	      x_min = x;
	    if ( x > x_max )
	      x_max = x;
	    if ( y < y_min )
	      y_min = y;
	    if ( y > y_max )
	      y_max = y;
	  }
	}
      }

      unlock(fr);
    }
    break;

  case 3 :
    if ( lock(fr) ) {
      for ( int y = r.y() ; y < r.y() + r.h() ; ++y ) {
	Uint8 *row = (Uint8 *)_s->pixels + y * _s->pitch + r.x() * 3;

	for ( int x = r.x() ; x < r.x() + r.w() ; ++x ) {
	  bool hit = false;

	  if ( *(row + _s->format->Ashift/8) != 0 )
	    hit = true;

	  row += 3;

	  if ( hit ) {
	    if ( x < x_min )
	      x_min = x;
	    if ( x > x_max )
	      x_max = x;
	    if ( y < y_min )
	      y_min = y;
	    if ( y > y_max )
	      y_max = y;
	  }
	}
      }

      unlock(fr);
    }
    break;

  case 4 :
    if ( lock(fr) ) {
      for ( int y = r.y() ; y < r.y() + r.h() ; ++y ) {
	Uint32 *row = (Uint32 *)((Uint8 *)_s->pixels + y * _s->pitch + r.x() * 4);

	for ( int x = r.x() ; x < r.x() + r.w() ; ++x ) {
	  bool hit = false;

	  if ( (*row++) & _s->format->Amask != 0 )
	    hit = true;

	  if ( hit ) {
	    if ( x < x_min )
	      x_min = x;
	    if ( x > x_max )
	      x_max = x;
	    if ( y < y_min )
	      y_min = y;
	    if ( y > y_max )
	      y_max = y;
	  }
	}
      }

      unlock(fr);
    }
    break;
  }

  return( rectangle(x_min, y_min, x_max - x_min + 1, y_max - y_min + 1) );
}

void surface::draw_pixel(Sint16 x, Sint16 y, Uint8 r, Uint8 g, Uint8 b) {
  SDL_Surface *_s = _surface[_current_frame];
  if ( !_s )
    return;

  if ( (x < _clip_box.x()) || (x >= _clip_box.x() + _clip_box.w()) ||
       (y < _clip_box.y()) || (y >= _clip_box.y() + _clip_box.h()) )
    return;

  if ( lock(_current_frame) ) {
    Uint32 pixel = SDL_MapRGB(_s->format, r, g, b);

    Uint8 bpp = _s->format->BytesPerPixel;
    Uint8 *bits = ((Uint8 *)_s->pixels) + y * _s->pitch + x * bpp;

    /* Set the pixel */
    switch(bpp) {
    case 1:
      *((Uint8 *)(bits)) = (Uint8)pixel;
      break;
    case 2:
      *((Uint16 *)(bits)) = (Uint16)pixel;
      break;
    case 3: { /* Format/endian independent */
      Uint8 r, g, b;
    
      r = (pixel>>_s->format->Rshift)&0xFF;
      g = (pixel>>_s->format->Gshift)&0xFF;
      b = (pixel>>_s->format->Bshift)&0xFF;
      *((bits)+_s->format->Rshift/8) = r; 
      *((bits)+_s->format->Gshift/8) = g;
      *((bits)+_s->format->Bshift/8) = b;
    }
    break;

    case 4:
      *((Uint32 *)(bits)) = (Uint32)pixel;
      break;
    }

    unlock(_current_frame);
  }
}

void surface::draw_pixels(Uint16 cnt, Sint16 *x, Sint16 *y, Uint8 *r, Uint8 *g, Uint8 *b) {
  SDL_Surface *_s = _surface[_current_frame];
  if ( !_s )
    return;

  if ( lock(_current_frame) ) {
    Uint8 bpp = _s->format->BytesPerPixel;

    switch ( bpp ) {
    case 1 :
      for ( Uint16 i = cnt ; i > 0 ; --i ) {
	if ( (*x >= _clip_box.x()) && (*x < _clip_box.x() + _clip_box.w()) &&
	     (*y >= _clip_box.y()) && (*y < _clip_box.y() + _clip_box.h()) ) {
	  Uint32 pixel = SDL_MapRGB(_s->format, *r, *g, *b);
	  Uint8 *bits = ((Uint8 *)_s->pixels) + (*y) * _s->pitch + (*x) * bpp;

	  *((Uint8 *)(bits)) = (Uint8)pixel;
	}

	x++; y++; r++; g++; b++;
      }
      break;

    case 2 :
      for ( Uint16 i = cnt ; i > 0 ; --i ) {
	if ( (*x >= _clip_box.x()) && (*x < _clip_box.x() + _clip_box.w()) &&
	     (*y >= _clip_box.y()) && (*y < _clip_box.y() + _clip_box.h()) ) {
	  Uint32 pixel = SDL_MapRGB(_s->format, *r, *g, *b);
	  Uint8 *bits = ((Uint8 *)_s->pixels) + (*y) * _s->pitch + (*x) * bpp;

	  *((Uint16 *)(bits)) = (Uint16)pixel;
	}

	x++; y++; r++; g++; b++;
      }
      break;

    case 3 : 
      for ( Uint16 i = cnt ; i > 0 ; --i ) {
	if ( (*x >= _clip_box.x()) && (*x < _clip_box.x() + _clip_box.w()) &&
	     (*y >= _clip_box.y()) && (*y < _clip_box.y() + _clip_box.h()) ) {
	  Uint32 pixel = SDL_MapRGB(_s->format, *r, *g, *b);
	  Uint8 *bits = ((Uint8 *)_s->pixels) + (*y) * _s->pitch + (*x) * bpp;

	  *((bits)+_s->format->Rshift/8) = (pixel>>_s->format->Rshift)&0xFF; 
	  *((bits)+_s->format->Gshift/8) = (pixel>>_s->format->Gshift)&0xFF;
	  *((bits)+_s->format->Bshift/8) = (pixel>>_s->format->Bshift)&0xFF;
	}

	x++; y++; r++; g++; b++;
      }
      break;

    case 4 :
      for ( Uint16 i = cnt ; i > 0 ; --i ) {
	if ( (*x >= _clip_box.x()) && (*x < _clip_box.x() + _clip_box.w()) &&
	     (*y >= _clip_box.y()) && (*y < _clip_box.y() + _clip_box.h()) ) {
	  Uint32 pixel = SDL_MapRGB(_s->format, *r, *g, *b);
	  Uint8 *bits = ((Uint8 *)_s->pixels) + (*y) * _s->pitch + (*x) * bpp;

	  *((Uint32 *)(bits)) = (Uint32)pixel;
	}

	x++; y++; r++; g++; b++;
      }
      break;
    }

    unlock(_current_frame);
  }
}

void surface::draw_rect(const rectangle &rect, Uint8 r, Uint8 g, Uint8 b) {
  SDL_Surface *_s = _surface[_current_frame];
  if ( !_s )
    return;

  SDL_FillRect(_s, (SDL_Rect *)rect.rect(), SDL_MapRGB(_s->format, r, g, b));
}

void surface::draw_text(Sint16 x, Sint16 y, bool solid, font *fnt, const char *txt) {
  while ( *txt ) {
    rectangle src = fnt->get_rect(*txt);
    if ( src.w() > 0 )
      copy(x + fnt->kern(*txt), y, fnt->image(), src);

    x += fnt->width(*txt);
    txt++;
  }
}

void surface::draw_mono_text(Sint16 x, Sint16 y, bool solid, Uint8, Uint8, Uint8, font *fnt, const char *txt) {
  draw_text(x, y, solid, fnt, txt);
}

void surface::draw_bob(Sint16 x, Sint16 y, Sint16 fr, surface *b) {
  rectangle bound = b->bound(fr);
  copy(x + bound.x(), y + bound.y(), b, fr, rectangle(0, 0, bound.w(), bound.h()));
}

void surface::copy(Sint16 x, Sint16 y, const surface *image, Uint16 fr, const rectangle &rect) {
  SDL_Surface *_s = _surface[_current_frame];
  if ( !_s )
    return;

  if ( (x + rect.w() >= _clip_box.x()) &&
       (x < _clip_box.x() + _clip_box.w()) &&
       (y + rect.h() >= _clip_box.y()) &&
       (y < _clip_box.y() + _clip_box.h()) ) {
    Sint16 src_xoff = rect.x();
    Sint16 src_width = rect.w();
    Sint16 src_yoff = rect.y();
    Sint16 src_height = rect.h();

    Sint16 dest_xoff = x;
    Sint16 dest_yoff = y;

    if ( src_xoff < 0 ) {
      src_width -= 0 - src_xoff;
      dest_xoff += 0 - src_xoff;
      src_xoff = 0;
    }

    if ( src_yoff < 0 ) {
      src_height -= 0 - src_yoff;
      dest_yoff += 0 - src_yoff;
      src_yoff = 0;
    }

    if ( src_xoff + src_width > image->_surface[fr]->w )
      src_width = image->_surface[fr]->w - src_xoff;

    if ( src_yoff + src_height > image->_surface[fr]->h )
      src_height = image->_surface[fr]->h - src_yoff;

    if ( dest_xoff < _clip_box.x() ) {
      src_xoff += _clip_box.x() - dest_xoff;
      src_width -= _clip_box.x() - dest_xoff;
      dest_xoff = _clip_box.x();
    }

    if ( dest_yoff < _clip_box.y() ) {
      src_yoff += _clip_box.y() - dest_yoff;
      src_height -= _clip_box.y() - dest_yoff;
      dest_yoff = _clip_box.x();
    }

    if ( dest_xoff + src_width > _clip_box.x() + _clip_box.w() )
      src_width = _clip_box.x() + _clip_box.w() - dest_xoff;
	
    if ( dest_yoff + src_height > _clip_box.y() + _clip_box.h() )
      src_height = _clip_box.y() + _clip_box.h() - dest_yoff;

    SDL_Rect src_r;
    src_r.x = src_xoff;
    src_r.y = src_yoff;
    src_r.w = src_width;
    src_r.h = src_height;
      
    SDL_Rect dest_r;
    dest_r.x = dest_xoff;
    dest_r.y = dest_yoff;
    dest_r.w = src_width;
    dest_r.h = src_height;
      
    SDL_BlitSurface(image->_surface[fr], &src_r, _s, &dest_r);
  }
}

bool surface::check_intersection(const surface *sf2, Uint16 fr1, Uint16 fr2, Sint16 x, Sint16 y) const {
  bool hit = true;
  const surface *sf1 = this;
  SDL_Surface *s1 = sf1->_surface[fr1];
  SDL_Surface *s2 = sf2->_surface[fr2];
  SDL_PixelFormat *f1 = s1->format;
  SDL_PixelFormat *f2 = s2->format;

  if ( f1->BytesPerPixel == f2->BytesPerPixel ) {
    hit = false;

    Sint16 l1 = sf1->_bound[fr1].x();
    Sint16 r1 = sf1->_bound[fr1].x() + sf1->_bound[fr1].w() - 1;
    Sint16 t1 = sf1->_bound[fr1].y();
    Sint16 b1 = sf1->_bound[fr1].y() + sf1->_bound[fr1].h() - 1;

    Sint16 l2 = x + sf2->_bound[fr2].x();
    Sint16 r2 = x + sf2->_bound[fr2].x() + sf2->_bound[fr2].w() - 1;
    Sint16 t2 = y + sf2->_bound[fr2].y();
    Sint16 b2 = y + sf2->_bound[fr2].y() + sf2->_bound[fr2].h() - 1;

    if ( (l2 <= r1) && (r2 >= l1) && (t2 <= b1) && (b2 >= t1) ) {
      Sint16 x1off, x2off, w;
      Sint16 y1off, y2off, h;

      if ( l2 >= l1 ) {
	x1off = l2 - l1;
	x2off = 0;
	w = r1 - l2 + 1;
      }
      else {
	x1off = 0;
	x2off = l1 - l2;
	w = r2 - l1 + 1;
      }

      if ( t2 >= t1 ) {
	y1off = t2 - t1;
	y2off = 0;
	h = b1 - t2 + 1;
      }
      else {
	y1off = 0;
	y2off = t1 - t2;
	h = b2 - t1 + 1;
      }

      switch ( s1->format->BytesPerPixel ) {
      case 1 : {
	// TODO
	hit = true;
	break;
      }

      case 2 : 
	if ( sf1->lock(fr1) ) {
	  if ( sf2->lock(fr2) ) {
	    for ( Sint16 ys = 0 ; !hit && (ys < h) ; ++ys ) {
	      Uint16 *p1 = (Uint16 *)(((Uint8 *)s1->pixels) + (y1off + ys) * s1->pitch + x1off * 2);
	      Uint16 *p2 = (Uint16 *)(((Uint8 *)s2->pixels) + (y2off + ys) * s2->pitch + x2off * 2);

	      for ( Sint16 xs = 0 ; xs < w ; ++xs ) {
		if ( (((*p1) & f1->Amask) >> f1->Ashift < 255) && (((*p2) & f2->Amask) >> f2->Ashift < 255) ) {
		  hit = true;
		  break;
		}

		++p1; ++p2;   
	      }
	    }

	    sf2->unlock(fr2);
	  }

	  sf1->unlock(fr1);
	}
	break;
      
      case 3 : {
	// TODO
	hit = true;
	break;
      }
      
      case 4 :
	if ( sf1->lock(fr1) ) {
	  if ( sf2->lock(fr2) ) {
	    for ( Sint16 ys = 0 ; !hit && (ys < h) ; ++ys ) {
	      Uint32 *p1 = (Uint32 *)(((Uint8 *)s1->pixels) + (y1off + ys) * s1->pitch + x1off * 4);
	      Uint32 *p2 = (Uint32 *)(((Uint8 *)s2->pixels) + (y2off + ys) * s2->pitch + x2off * 4);

	      for ( Sint16 xs = 0 ; xs < w ; ++xs ) {
		if ( (((*p1) & f1->Amask) >> f1->Ashift < 255) && (((*p2) & f2->Amask) >> f2->Ashift < 255) ) {
		  hit = true;
		  break;
		}

		++p1; ++p2;
	      }
	    }

	    sf2->unlock(fr2);
	  }

	  sf1->unlock(fr1);
	}
	break;
      }
    }
  }

  return( hit );
}

